local g3d = require("g3d")

local lightShader = love.graphics.newShader(g3d.shaderpath, [[

    // https://love2d.org/forums/viewtopic.php?p=244728#p244728

    varying vec4 worldPosition;
    varying vec3 vertexNormal;

    uniform mat4 modelMatrix;
    uniform vec3 lightPosition;
    uniform float ambient = 0.2;

    vec4 effect(vec4 color, Image tex, vec2 texcoord, vec2 pixcoord) {

        vec3 lightDirection = normalize(lightPosition.xyz - worldPosition.xyz);
        vec3 normal = normalize(mat3(modelMatrix) * vertexNormal);

        float diffuse = max(dot(lightDirection, normal), 0);

        vec4 texcolor = Texel(tex, texcoord);
        if (texcolor.a == 0.0) { discard; }

        float lightness = diffuse + ambient;
        return vec4((texcolor * color).rgb * lightness, 1.0);

    }

]])

function newCylinder(height, radius, nb_segments)

    local vertices = {}
    local angle_step = 2 * math.pi / nb_segments
    local angle = 0

    local base_z = -height / 2
    local tip_z = height / 2

    for i = 1, nb_segments do

        local x1 = radius * math.cos(angle)
        local y1 = radius * math.sin(angle)
        local x2 = radius * math.cos(angle + angle_step)
        local y2 = radius * math.sin(angle + angle_step)

        -- Base triangle

        table.insert(vertices, {0,0,base_z, 0.5,0.5, 0,0,-1})
        table.insert(vertices, {x1,y1,base_z, 0,1, 0,0,-1})
        table.insert(vertices, {x2,y2,base_z, 1,1, 0,0,-1})

        -- Tip triangle

        table.insert(vertices, {0,0,tip_z, 0.5,0.5, 0,0,1})
        table.insert(vertices, {x1,y1,tip_z, 0,0, 0,0,1})
        table.insert(vertices, {x2,y2,tip_z, 1,0, 0,0,1})

        -- Side triangles

        local edge1_x, edge1_y, edge1_z = x1 - x2, y1 - y2, base_z - tip_z
        local edge2_x, edge2_y, edge2_z = x2 - x1, y2 - y1, base_z - tip_z

        local normal_x = edge1_y * edge2_z - edge1_z * edge2_y
        local normal_y = edge1_z * edge2_x - edge1_x * edge2_z
        local normal_z = edge1_x * edge2_y - edge1_y * edge2_x

        local normal_length = math.sqrt(normal_x * normal_x + normal_y * normal_y + normal_z * normal_z)

        nx1, ny1, nz1 = normal_x / normal_length, normal_y / normal_length, normal_z / normal_length
        nx2, ny2, nz2 = normal_x / normal_length, normal_y / normal_length, normal_z / normal_length

        table.insert(vertices, {x1,y1,base_z, 0, 0.5 - 0.5*y1/radius, nx1,ny1,nz1})
        table.insert(vertices, {x1,y1,tip_z, 1, 0.5 - 0.5*y1/radius, nx1,ny1,nz1})
        table.insert(vertices, {x2,y2,base_z, 0, 0.5 - 0.5*y2/radius, nx2,ny2,nz2})

        table.insert(vertices, {x2,y2,tip_z, 1, 0.5 - 0.5*y2/radius, nx2,ny2,nz2})
        table.insert(vertices, {x2,y2,base_z, 0, 0.5 - 0.5*y2/radius, nx2,ny2,nz2})
        table.insert(vertices, {x1,y1,tip_z, 1, 0.5 - 0.5*y1/radius, nx1,ny1,nz1})

        angle = angle + angle_step

    end

    return vertices

end


function love.load()
    CylinderLine = g3d.newModel(
        newCylinder(10, .01, 256), nil,
        {0,0,0}, {math.pi/2,math.pi/2,0}
    )
end

function love.update(dt)
    g3d.camera.firstPersonMovement(dt)
    lightShader:send("lightPosition", g3d.camera.position)
end

function love.mousemoved(x, y, dx, dy)

    local cameraX, cameraY, cameraZ = unpack(g3d.camera.position)
    local width, height = love.graphics.getDimensions()

    local directionX = (height / 2) / math.tan(g3d.camera.fov / 2) 
    local directionY = x - width / 2
    local directionZ = y - height / 2

    local length = -math.sqrt(directionX^2 + directionY^2 + directionZ^2)
    directionX, directionY, directionZ = directionX/length, directionY/length, directionZ/length

    local cylinderX, cylinderY, cylinderZ = cameraX + 10 + directionX, cameraY + directionY, cameraZ + directionZ
    local cylinderTipX, cylinderTipY, cylinderTipZ = cylinderX, cylinderY, cylinderZ + 10

    CylinderLine:setTranslation(cylinderX, cylinderY, cylinderZ)

    CylinderLine:setRotation(
        math.atan2(cylinderTipZ - cylinderZ, cylinderTipX - cylinderX),
        0,
        math.atan2(math.sqrt((cylinderTipX - cylinderX)^2 + (cylinderTipZ - cylinderZ)^2), cylinderTipY - cylinderY)
    )

    local dist = math.sqrt((cylinderX - cameraX)^2 + (cylinderY - cameraY)^2 + (cylinderZ - cameraZ)^2)

    CylinderLine:setScale(1, 1, dist/5.5)

end

function love.draw()
    love.graphics.setColor(1,1,0)
    CylinderLine:draw(lightShader)
end


